from typing import TYPE_CHECKING, Any, Dict, Optional, Tuple, Type, Union, get_type_hints

from pydantic import BaseModel

from pait.util import create_pydantic_model

if TYPE_CHECKING:
    from pydantic.typing import AbstractSetIntStr, DictStrAny, MappingIntStrAny


class PaitBaseModel(object):
    """pait base model, The value of the attribute supports filling in the pait field
    >>> from pait import field
    >>> class Demo(PaitBaseModel):
    ...     a: int = field.Query.i()
    ...     b: str = field.Body.i()
    """

    _pydantic_model: Optional[Type[BaseModel]] = None

    @classmethod
    def to_pydantic_model(cls) -> Type[BaseModel]:
        if cls._pydantic_model is not None:
            return cls._pydantic_model
        else:
            annotation_dict: Dict[str, Tuple[Type, Any]] = {
                param_name: (annotation, getattr(cls, param_name, ...))
                for param_name, annotation in get_type_hints(cls).items()
                if not param_name.startswith("_")
            }
            cls._pydantic_model = create_pydantic_model(annotation_dict)
            return cls._pydantic_model

    def dict(
        self,
        include: Union["AbstractSetIntStr", "MappingIntStrAny"] = None,
        exclude: Union["AbstractSetIntStr", "MappingIntStrAny"] = None,
        by_alias: bool = False,
        skip_defaults: bool = None,
        exclude_unset: bool = False,
        exclude_defaults: bool = False,
        exclude_none: bool = False,
    ) -> "DictStrAny":
        if self._pydantic_model is None:
            raise NotImplementedError
        _pydantic_model: BaseModel = self._pydantic_model(**self.__dict__)
        return _pydantic_model.dict(
            include=include,
            exclude=exclude,
            by_alias=by_alias,
            skip_defaults=skip_defaults,
            exclude_unset=exclude_unset,
            exclude_defaults=exclude_defaults,
            exclude_none=exclude_none,
        )

    @classmethod
    def schema(cls, by_alias: bool = True) -> "DictStrAny":
        return cls.to_pydantic_model().schema(by_alias=by_alias)



import copy
import json
from typing import Any, Dict, List, Optional, Tuple, Type

import yaml  # type: ignore
from pydantic import BaseModel, Field, HttpUrl
from pydantic.fields import Undefined

from pait import field as pait_field
from pait.api_doc.base_parse import PaitBaseParse
from pait.g import config
from pait.model.core import PaitCoreModel
from pait.model.response import PaitResponseModel
from pait.model.status import PaitStatus
from pait.util import create_pydantic_model

from .base_parse import FieldDictType, FieldSchemaTypeDict

__all__ = ["PaitOpenApi"]


class _OpenApiInfoModel(BaseModel):
    """open api info column model"""

    class _Contact(BaseModel):
        name: str
        url: str
        email: str

    class _License(BaseModel):
        name: str
        url: str

    title: str = Field("Pait Open Api")
    description: str = Field(None)
    version: str = Field("0.0.1")
    contact: _Contact = Field(None)
    license: _License = Field(None)


class _OpenApiTagModel(BaseModel):
    """openapi tag column model"""

    class _ExternalDocs(BaseModel):
        url: HttpUrl

    name: str
    description: str = Field(None)
    externalDocs: str = Field(None)


class _OpenApiServerModel(BaseModel):
    """openapi server column model"""

    url: HttpUrl = Field("http://127.0.0.1")
    description: str = Field(None)


class PaitOpenApi(PaitBaseParse):
    def __init__(
        self,
        pait_dict: Dict[str, PaitCoreModel],
        title: Optional[str] = None,
        open_api_info: Optional[Dict[str, Any]] = None,
        open_api_tag_list: Optional[List[Dict[str, Any]]] = None,
        open_api_server_list: Optional[List[Dict[str, Any]]] = None,
        # default_response: Optional[...] = None,  # TODO
        type_: str = "json",
    ):
        super().__init__(pait_dict)
        self._header_keyword_dict: Dict[str, str] = {
            "Content-Type": "requestBody.content.<media-type>",
            "Accept": "responses.<code>.content.<media-type>",
            "Authorization": " security",
        }

        if not open_api_info:
            open_api_info = _OpenApiInfoModel(title=title).dict(exclude_none=True)
        else:
            open_api_info = _OpenApiInfoModel(**open_api_info).dict()

        if not open_api_server_list:
            open_api_server_list = [_OpenApiServerModel().dict(exclude_none=True)]
        else:
            temp_open_api_server_list: List[Dict[str, Any]] = []
            for open_api_server in open_api_server_list:
                temp_open_api_server_list.append(_OpenApiServerModel(**open_api_server).dict(exclude_none=True))
            open_api_server_list = temp_open_api_server_list

        if not open_api_tag_list:
            open_api_tag_list = []
        else:
            temp_open_api_tag_list: List[Dict[str, Any]] = []
            for open_api_tag in open_api_tag_list:
                temp_open_api_tag_list.append(_OpenApiTagModel(**open_api_tag).dict(exclude_none=True))
            open_api_tag_list = temp_open_api_tag_list

        self.open_api_dict: Dict[str, Any] = {
            "openapi": "3.0.0",
            "info": open_api_info,
            "servers": open_api_server_list,
            "tags": open_api_tag_list,
            "paths": {},
            "components": {"schemas": {}},
            # TODO
            # "security": {},
            # "externalDocs": {}
        }
        self.parse_data_2_openapi()
        if type_ == "json":
            self.content = json.dumps(self.open_api_dict, cls=config.json_encoder)
            self._content_type = ".json"
        elif type_ == "yaml":
            self.content = yaml.dump(self.open_api_dict, sort_keys=False)
            self._content_type = ".yaml"

    def _replace_pydantic_definitions(self, schema: dict, path: str, parent_schema: Optional[dict] = None) -> None:
        """update schemas'definitions to components schemas"""
        if not parent_schema:
            parent_schema = schema
        for key, value in schema.items():
            if key == "$ref" and not value.startswith("#/components"):
                index: int = value.rfind("/") + 1
                model_key: str = value[index:]
                schema[key] = f"#/components/schemas/{model_key}"
                self.open_api_dict["components"]["schemas"][model_key] = parent_schema["definitions"][model_key]
            elif isinstance(value, dict):
                self._replace_pydantic_definitions(value, path, parent_schema)
            elif isinstance(value, list):
                for item in value:
                    if isinstance(item, dict):
                        self._replace_pydantic_definitions(item, path, parent_schema)

    @staticmethod
    def _snake_name_to_hump_name(name: str) -> str:
        return "".join([item.upper() for item in name.split("_")])

    def _field_list_2_request_body(
        self,
        media_type: str,
        openapi_method_dict: dict,
        field_dict_list: List[FieldSchemaTypeDict],
        operation_id: str,
    ) -> None:
        """gen request body schema and update request body schemas'definitions to components schemas"""
        openapi_request_body_dict: dict = openapi_method_dict.setdefault("requestBody", {"content": {}})
        annotation_dict: Dict[str, Tuple[Type, pait_field.BaseField]] = {
            field_dict["raw"]["param_name"]: (field_dict["raw"]["annotation"], field_dict["raw"]["field"])
            for field_dict in field_dict_list
        }
        _pydantic_model: Type[BaseModel] = create_pydantic_model(
            annotation_dict, class_name=f"{self._snake_name_to_hump_name(operation_id)}DynamicModel"
        )
        schema_dict = copy.deepcopy(_pydantic_model.schema())
        path = f"#/components/schemas/{schema_dict['title']}"
        self._replace_pydantic_definitions(schema_dict, path)
        if "definitions" in schema_dict:
            del schema_dict["definitions"]

        openapi_request_body_dict["content"].update({media_type: {"schema": schema_dict}})

    # flake8: noqa: C901
    def parse_data_2_openapi(self) -> None:
        for group, pait_model_list in self._group_pait_dict.items():
            for pait_model in pait_model_list:
                path: str = pait_model.openapi_path
                openapi_path_dict: dict = self.open_api_dict["paths"].setdefault(path, {})
                for method in pait_model.method_list:
                    openapi_method_dict: dict = openapi_path_dict.setdefault(method.lower(), {})
                    if pait_model.tag:
                        openapi_method_dict["tags"] = list(pait_model.tag)
                        for tag in pait_model.tag:
                            if tag not in {tag_dict["name"] for tag_dict in self.open_api_dict["tags"]}:
                                self.open_api_dict["tags"].append(
                                    {
                                        "name": tag,
                                        "description": "",
                                    }
                                )
                    if pait_model.status in (
                        PaitStatus.abnormal,
                        PaitStatus.maintenance,
                        PaitStatus.archive,
                        PaitStatus.abandoned,
                    ):
                        openapi_method_dict["deprecated"] = True
                    openapi_method_dict["summary"] = pait_model.summary
                    openapi_method_dict["description"] = pait_model.desc
                    openapi_method_dict["operationId"] = f"{method}.{pait_model.operation_id}"
                    openapi_parameters_list: list = openapi_method_dict.setdefault("parameters", [])
                    openapi_response_dict: dict = openapi_method_dict.setdefault("responses", {})
                    all_field_dict: FieldDictType = self._parse_func_param_to_field_dict(pait_model.func)
                    for pre_depend in pait_model.pre_depend_list:
                        for field_class, field_dict_list in self._parse_func_param_to_field_dict(pre_depend).items():
                            if field_class not in all_field_dict:
                                all_field_dict[field_class] = field_dict_list
                            else:
                                all_field_dict[field_class].extend(field_dict_list)

                    for field_class, field_dict_list in all_field_dict.items():
                        if field_class in (
                            pait_field.Cookie,
                            pait_field.Header,
                            pait_field.Path,
                            pait_field.Query,
                        ):
                            for field_dict in field_dict_list:
                                param_name: str = field_dict["raw"]["param_name"]
                                if field_class == pait_field.Header:
                                    param_name = self._header_keyword_dict.get(param_name, param_name)
                                # TODO support example
                                openapi_parameters_list.append(
                                    {
                                        "name": param_name,
                                        "in": field_class.cls_lower_name(),
                                        "required": field_dict["default"] is Undefined,
                                        # openapi description must not null
                                        "description": field_dict["description"] or "",
                                        "schema": field_dict["raw"]["schema"],
                                    }
                                )
                        elif field_class == pait_field.Body:
                            # support args BodyField
                            self._field_list_2_request_body(
                                "application/json", openapi_method_dict, field_dict_list, pait_model.operation_id
                            )
                        elif field_class == pait_field.Form:
                            # support args FormField
                            self._field_list_2_request_body(
                                "application/x-www-form-urlencoded",
                                openapi_method_dict,
                                field_dict_list,
                                pait_model.operation_id,
                            )
                        else:
                            # TODO
                            pass

                    if pait_model.response_model_list:
                        response_schema_dict: Dict[tuple, List[Dict[str, str]]] = {}
                        for resp_model_class in pait_model.response_model_list:
                            resp_model: PaitResponseModel = resp_model_class()
                            schema_dict: dict = {}
                            if resp_model.response_data:
                                schema_dict = resp_model.response_data.schema()

                            # fix del schema dict
                            schema_dict = copy.deepcopy(schema_dict)
                            path = f"#/components/schemas/{schema_dict['title']}"
                            self._replace_pydantic_definitions(schema_dict, path)
                            if "definitions" in schema_dict:
                                del schema_dict["definitions"]
                            for _status_code in resp_model.status_code:
                                key: tuple = (_status_code, resp_model.media_type)
                                ref_dict: dict = {"$ref": path}
                                if key in response_schema_dict:
                                    response_schema_dict[key].append(ref_dict)
                                else:
                                    response_schema_dict[key] = [ref_dict]
                                if _status_code in openapi_response_dict:
                                    openapi_response_dict[_status_code]["description"] += f"|{resp_model.description}"
                                else:
                                    openapi_response_dict[_status_code] = {"description": resp_model.description}
                                self.open_api_dict["components"]["schemas"].update({schema_dict["title"]: schema_dict})
                        # mutli response support
                        # only response example see https://swagger.io/docs/specification/describing-responses/   FAQ
                        for key_tuple, path_list in response_schema_dict.items():
                            status_code, media_type = key_tuple
                            if len(path_list) == 1:
                                ref_dict = path_list[0]
                            else:
                                ref_dict = {"oneOf": path_list}
                            openapi_response_dict[status_code]["content"] = {media_type: {"schema": ref_dict}}



from ._func_sig import FuncSig, get_func_sig
from ._lazy_property import LazyProperty
from ._util import (
    Undefined,
    UndefinedType,
    create_pydantic_model,
    gen_example_dict_from_schema,
    gen_example_json_from_python,
    gen_example_json_from_schema,
    gen_tip_exc,
    get_parameter_list_from_class,
)



import logging
from typing import Callable, Dict, Optional, Set, Type

from sanic.app import Sanic
from sanic_testing.testing import SanicTestClient, TestingResponse  # type: ignore

from pait.g import pait_data
from pait.model.core import PaitCoreModel

from ._app_helper import AppHelper

__all__ = ["load_app"]


def load_app(app: Sanic, project_name: str = "") -> Dict[str, PaitCoreModel]:
    """Read data from the route that has been registered to `pait`"""
    _pait_data: Dict[str, PaitCoreModel] = {}
    for route in app.router.routes:
        if "static" in route.name:
            continue

        route_name: str = route.name
        method_set: Set[str] = route.methods
        path: str = route.path
        openapi_path: str = path
        handler: Callable = route.handler

        # replace path <xxx> to {xxx}
        if "<" in openapi_path and ">" in openapi_path:
            new_path_list: list = []
            for sub_path in openapi_path.split("/"):
                if sub_path[0] == "<" and sub_path[-1] == ">":
                    real_sub_path: str = sub_path[1:-1]
                    if ":" in real_sub_path:
                        real_sub_path = real_sub_path.split(":")[0]
                    real_sub_path = "{" + real_sub_path + "}"
                else:
                    real_sub_path = sub_path
                new_path_list.append(real_sub_path)
            openapi_path = "/".join(new_path_list)
        if not openapi_path.startswith("/"):
            openapi_path = "/" + openapi_path
        for method in method_set:
            view_class: Optional[Type] = getattr(handler, "view_class", None)
            if view_class:
                handler = getattr(view_class, method.lower(), None)
            if not handler:
                logging.warning(f"{route_name} can not found handle")
                continue
            pait_id: Optional[str] = getattr(handler, "_pait_id", None)
            if not pait_id:
                logging.warning(f"{route_name} can not found pait id")
                continue
            pait_data.add_route_info(
                AppHelper.app_name, pait_id, path, openapi_path, {method}, route_name, project_name
            )
            _pait_data[pait_id] = pait_data.get_pait_data(AppHelper.app_name, pait_id)

        # old version
        # for path, handler_dict in route.handlers.items():
        #     for method, handler_list in handler_dict.items():
        #         for handler in handler_list:
        #             view_class: Optional[Type] = getattr(handler, "view_class", None)
        #             if view_class:
        #                 handler = getattr(view_class, method.lower(), None)
        #             if not handler:
        #                 logging.warning(f"{route_name} can not found handle")
        #                 continue
        #             pait_id: Optional[str] = getattr(handler, "_pait_id", None)
        #             if not pait_id:
        #                 logging.warning(f"{route_name} can not found pait id")
        #                 continue
        #             pait_data.add_route_info(AppHelper.app_name, pait_id, path, {method}, route_name, project_name)
        #             _pait_data[pait_id] = pait_data.get_pait_data(AppHelper.app_name, pait_id)
    return _pait_data



from typing import Any, Dict, List, Mapping, Tuple

from sanic.headers import HeaderIterable
from sanic.request import File, Request, RequestParameters
from sanic_testing.testing import SanicTestClient, TestingResponse  # type: ignore

from pait.app.base import BaseAppHelper
from pait.util import LazyProperty

__all__ = ["AppHelper"]


class AppHelper(BaseAppHelper):
    RequestType = Request
    FormType = RequestParameters
    FileType = File
    HeaderType = HeaderIterable
    app_name = "sanic"

    def __init__(self, class_: Any, args: Tuple[Any, ...], kwargs: Mapping[str, Any]):
        super().__init__(class_, args, kwargs)

    def body(self) -> dict:
        return self.request.json

    def cookie(self) -> dict:
        return self.request.cookies

    def file(self) -> Dict[str, File]:
        return {key: value[0] for key, value in self.request.files.items()}

    def form(self) -> dict:
        return {key: value[0] for key, value in self.request.form.items()}

    def header(self) -> HeaderIterable:
        return self.request.headers

    def path(self) -> Mapping[str, Any]:
        return self.request_kwargs

    @LazyProperty()
    def query(self) -> dict:
        return {key: value[0] for key, value in self.request.args.items()}

    @LazyProperty()
    def multiform(self) -> Dict[str, List[Any]]:
        return {key: self.request.form.getlist(key) for key, _ in self.request.form.items()}

    @LazyProperty()
    def multiquery(self) -> Dict[str, Any]:
        return {key: self.request.args.getlist(key) for key, _ in self.request.args.items()}



import sys
from typing import Any, List

app_list: List = ["flask", "starlette", "sanic", "tornado"]


def auto_load_app_class() -> Any:
    """A project using only a web framework, to use `auto_load_app_class`"""
    real_app: Any = None
    for app in app_list:
        if app not in sys.modules:
            continue
        if real_app:
            raise RuntimeError(f"Pait unable to make a choice {real_app} & {app}")
        real_app = sys.modules[app]
    if not real_app:
        raise RuntimeError("Pait can't auto load app class")
    return real_app



from importlib import import_module
from typing import Any, Callable, Dict, List, Optional, Tuple, Type

from pait.model.status import PaitStatus

from ..model.core import PaitCoreModel  # type: ignore
from ..model.response import PaitResponseModel  # type: ignore
from .auto_load_app import auto_load_app_class  # type: ignore


def _import_func_from_app(app: Any, fun_name: str) -> Callable:
    app_name: str = app.__class__.__name__.lower()
    if app_name == "flask":
        return getattr(import_module("pait.app.flask"), fun_name)
    elif app_name == "starlette":
        return getattr(import_module("pait.app.starlette"), fun_name)
    elif app_name == "sanic":
        return getattr(import_module("pait.app.sanic"), fun_name)
    elif app_name == "application" and app.__class__.__module__ == "tornado.web":
        return getattr(import_module("pait.app.tornado"), fun_name)
    else:
        raise NotImplementedError(f"Pait not support app name:{app_name}, please check app:{app}")


def _base_call_func(app: Any, fun_name: str, *args: Any, **kwargs: Any) -> Any:
    return _import_func_from_app(app, fun_name)(*args, **kwargs)


def load_app(app: Any, project_name: str = "") -> Dict[str, PaitCoreModel]:
    """Read data from the route that has been registered to `pait`
    Note:This is an implicit method
    """
    return _base_call_func(app, "load_app", app, project_name=project_name)


def add_doc_route(app: Any, prefix: str = "/") -> None:
    return _base_call_func(app, "add_doc_route", prefix=prefix)


def pait(
    # param check
    at_most_one_of_list: Optional[List[List[str]]] = None,
    required_by: Optional[Dict[str, List[str]]] = None,
    # doc
    author: Optional[Tuple[str]] = None,
    desc: Optional[str] = None,
    summary: Optional[str] = None,
    name: Optional[str] = None,
    status: Optional[PaitStatus] = None,
    group: Optional[str] = None,
    tag: Optional[Tuple[str, ...]] = None,
    response_model_list: Optional[List[Type[PaitResponseModel]]] = None,
) -> Callable:
    """provide parameter checks and type conversions for each routing function/cbv class
    Note:This is an implicit method
    """
    load_class_app = auto_load_app_class()
    app_name: str = load_class_app.__name__.lower()
    if app_name == "flask":
        from .flask import pait as _pait  # type: ignore
    elif app_name == "starlette":
        from .starlette import pait as _pait  # type: ignore
    elif app_name == "sanic":
        from .sanic import pait as _pait  # type: ignore
    elif app_name == "tornado":
        from .tornado import pait as _pait  # type: ignore
    else:
        raise NotImplementedError(f"Pait not support:{load_class_app}")
    return _pait(
        at_most_one_of_list=at_most_one_of_list,
        required_by=required_by,
        author=author,
        desc=desc,
        summary=summary,
        name=name,
        status=status,
        group=group,
        tag=tag,
        response_model_list=response_model_list,
    )



from typing import Any, Callable, Dict, List, Optional, Tuple, Type

from pydantic import BaseConfig
from tornado.web import RequestHandler

from pait.core import pait as _pait
from pait.model.response import PaitResponseModel
from pait.model.status import PaitStatus
from pait.util import gen_example_json_from_schema

from ._app_helper import AppHelper

__all__ = ["pait"]


def make_mock_response(pait_response: Type[PaitResponseModel]) -> Any:
    tornado_handle: RequestHandler = getattr(pait_response, "handle", None)
    if not tornado_handle:
        raise RuntimeError("Can not load Tornado handle")
    tornado_handle.set_status(pait_response.status_code[0])
    for key, value in pait_response.header.items():
        tornado_handle.set_header(key, value)
    if pait_response.media_type == "application/json" and pait_response.response_data:
        tornado_handle.write(gen_example_json_from_schema(pait_response.response_data.schema(), use_example_value=True))
        return
    else:
        raise NotImplementedError()


def pait(
    # param check
    at_most_one_of_list: Optional[List[List[str]]] = None,
    required_by: Optional[Dict[str, List[str]]] = None,
    pre_depend_list: Optional[List[Callable]] = None,
    # doc
    author: Optional[Tuple[str]] = None,
    desc: Optional[str] = None,
    summary: Optional[str] = None,
    name: Optional[str] = None,
    status: Optional[PaitStatus] = None,
    group: Optional[str] = None,
    tag: Optional[Tuple[str, ...]] = None,
    enable_mock_response: bool = False,
    response_model_list: Optional[List[Type[PaitResponseModel]]] = None,
    pydantic_model_config: Optional[Type[BaseConfig]] = None,
) -> Callable:
    """Help starlette provide parameter checks and type conversions for each routing function/cbv class"""
    return _pait(
        AppHelper,
        make_mock_response_fn=make_mock_response,
        author=author,
        desc=desc,
        summary=summary,
        name=name,
        status=status,
        group=group,
        tag=tag,
        enable_mock_response=enable_mock_response,
        response_model_list=response_model_list,
        pre_depend_list=pre_depend_list,
        at_most_one_of_list=at_most_one_of_list,
        required_by=required_by,
        pydantic_model_config=pydantic_model_config,
    )



from dataclasses import MISSING
from typing import Any, Callable, Optional

from pydantic.fields import FieldInfo, Undefined
from pydantic.typing import NoArgAnyCallable


class BaseField(FieldInfo):
    def __init__(
        self,
        default: Any = Undefined,
        *,
        raw_return: bool = False,
        example: Any = MISSING,
        default_factory: Optional[NoArgAnyCallable] = None,
        alias: str = None,
        title: str = None,
        description: str = None,
        const: bool = None,
        gt: float = None,
        ge: float = None,
        lt: float = None,
        le: float = None,
        multiple_of: float = None,
        min_items: int = None,
        max_items: int = None,
        min_length: int = None,
        max_length: int = None,
        regex: str = None,
        **extra: Any,
    ):
        if self.__class__.__mro__[2] != FieldInfo:
            raise RuntimeError("Only classes that inherit BaseField can be used")
        self.raw_return = raw_return
        self.lower_name: str = self.__class__.__name__.lower()
        if example is not MISSING:
            extra["example"] = example
        super().__init__(
            default,
            default_factory=default_factory,
            alias=alias,
            title=title,
            description=description,
            const=const,
            gt=gt,
            ge=ge,
            lt=lt,
            le=le,
            multiple_of=multiple_of,
            min_items=min_items,
            max_items=max_items,
            min_length=min_length,
            max_length=max_length,
            regex=regex,
            **extra,
        )

    @classmethod
    def cls_lower_name(cls) -> str:
        return cls.__name__.lower()

    def __lt__(self, other: "BaseField") -> bool:
        return self.lower_name < other.lower_name

    @classmethod
    def i(
        cls,
        default: Any = Undefined,
        *,
        raw_return: bool = False,
        example: Any = MISSING,
        default_factory: Optional[NoArgAnyCallable] = None,
        alias: str = None,
        title: str = None,
        description: str = None,
        const: bool = None,
        gt: float = None,
        ge: float = None,
        lt: float = None,
        le: float = None,
        multiple_of: float = None,
        min_items: int = None,
        max_items: int = None,
        min_length: int = None,
        max_length: int = None,
        regex: str = None,
        **extra: Any,
    ) -> Any:
        """ignore mypy tip"""
        return cls(
            default,
            raw_return=raw_return,
            example=example,
            default_factory=default_factory,
            alias=alias,
            title=title,
            description=description,
            const=const,
            gt=gt,
            ge=ge,
            lt=lt,
            le=le,
            multiple_of=multiple_of,
            min_items=min_items,
            max_items=max_items,
            min_length=min_length,
            max_length=max_length,
            regex=regex,
            **extra,
        )


class Body(BaseField):
    pass


class Cookie(BaseField):
    pass


class File(BaseField):
    pass


class Form(BaseField):
    pass


class MultiForm(BaseField):
    pass


class Header(BaseField):
    pass


class Path(BaseField):
    pass


class Query(BaseField):
    pass


class MultiQuery(BaseField):
    pass


class Depends(object):
    def __init__(self, func: Callable):
        self.func: Callable = func

    @classmethod
    def i(cls, func: Callable) -> Any:
        return cls(func)



import logging
from typing import TYPE_CHECKING, Dict, Set

if TYPE_CHECKING:
    from pait.model.core import PaitCoreModel


class PaitData(object):
    def __init__(self) -> None:
        self.pait_id_dict: Dict[str, Dict[str, "PaitCoreModel"]] = {}

    def register(self, app_name: str, pait_info_model: "PaitCoreModel") -> None:
        """Store the data of each routing handle"""
        pait_id: str = pait_info_model.pait_id
        if app_name not in self.pait_id_dict:
            self.pait_id_dict[app_name] = {}
        self.pait_id_dict[app_name][pait_id] = pait_info_model

    def get_pait_data(self, app_name: str, pait_id: str) -> "PaitCoreModel":
        """Get route handle data"""
        return self.pait_id_dict[app_name][pait_id]

    def add_route_info(
        self,
        app_name: str,
        pait_id: str,
        path: str,
        openapi_path: str,
        method_set: Set[str],
        route_name: str,
        project_name: str,
    ) -> None:
        """Route handle information supplemented by load_app"""
        if pait_id in self.pait_id_dict[app_name]:
            model: "PaitCoreModel" = self.pait_id_dict[app_name][pait_id]
            model.path = path
            model.openapi_path = openapi_path
            model.method_list = sorted(list(method_set or set()), reverse=True)
            model.operation_id = route_name
            if project_name:
                model.func_path = project_name + f"{project_name}".join(
                    model.func.__code__.co_filename.split(project_name)[1:]
                )
            else:
                model.func_path = model.func.__code__.co_filename
        else:
            logging.warning(f"load path:{path} fail, pait id:{pait_id}")

    def __bool__(self) -> bool:
        return bool(self.pait_id_dict)



from typing import Any, Dict, List, Mapping, Tuple

from flask import Request, request
from werkzeug.datastructures import EnvironHeaders, ImmutableMultiDict

from pait.app.base import BaseAppHelper
from pait.util import LazyProperty

__all__ = ["AppHelper"]


class AppHelper(BaseAppHelper):
    RequestType = Request
    FormType = ImmutableMultiDict
    FileType = Request.files
    HeaderType = EnvironHeaders
    app_name = "flask"

    def __init__(self, class_: Any, args: Tuple[Any, ...], kwargs: Mapping[str, Any]):
        super().__init__(class_, args, kwargs)

        self.request = request

    def body(self) -> dict:
        return request.json

    def cookie(self) -> dict:
        return request.cookies

    def file(self) -> Request.files:  # type: ignore
        return request.files

    def form(self) -> Request.form:  # type: ignore
        return request.form

    def header(self) -> EnvironHeaders:
        return request.headers

    def path(self) -> Mapping[str, Any]:
        return self.request_kwargs

    def query(self) -> Dict[str, Any]:
        return request.args

    @LazyProperty()
    def multiform(self) -> Dict[str, List[Any]]:
        return {key: request.form.getlist(key) for key, _ in request.form.items()}

    @LazyProperty()
    def multiquery(self) -> Dict[str, List[Any]]:
        return {key: request.args.getlist(key) for key, _ in request.args.items()}



def get_redoc_html(
    open_api_json_url: str,
    title: str = "ReDoc",
) -> str:
    """copy from https://github.com/Redocly/redoc#tldr"""
    return f"""
    <!DOCTYPE html>
    <html>
      <head>
        <title>{title}</title>
        <!-- needed for adaptive design -->
        <meta charset="utf-8"/>
        <meta name="viewport" content="width=device-width, initial-scale=1">
        <link href="https://fonts.googleapis.com/css?family=Montserrat:300,400,700|Roboto:300,400,700" rel="stylesheet">

        <!--
        ReDoc doesn't change outer page styles
        -->
        <style>
          body {{
            margin: 0;
            padding: 0;
          }}
        </style>
      </head>
      <body>
        <redoc spec-url='{open_api_json_url}'></redoc>
        <script src="https://cdn.jsdelivr.net/npm/redoc@next/bundles/redoc.standalone.js"> </script>
      </body>
    </html>
    """


def get_swagger_ui_html(
    open_api_json_url: str,
    title: str = "ReDoc",
) -> str:

    return f"""
    <!DOCTYPE html>
    <html>
    <head>
    <link type="text/css" rel="stylesheet" href="https://cdn.jsdelivr.net/npm/swagger-ui-dist@3/swagger-ui.css">
    <link rel="shortcut icon" href="https://static1.smartbear.co/swagger/media/assets/swagger_fav.png">
    <title>{title}</title>
    </head>
    <body>
    <div id="swagger-ui">
    </div>
    <script src="https://cdn.jsdelivr.net/npm/swagger-ui-dist@3/swagger-ui-bundle.js"></script>
    <!-- `SwaggerUIBundle` is now available on the page -->
    <script>
    const ui = SwaggerUIBundle({{
        url: '{open_api_json_url}',
        dom_id: '#swagger-ui',
        presets: [
        SwaggerUIBundle.presets.apis,
        SwaggerUIBundle.SwaggerUIStandalonePreset
        ],
        layout: "BaseLayout",
        deepLinking: true,
        showExtensions: true,
        showCommonExtensions: true
    }})
    </script>
    </body>
    </html>
    """



from typing import Dict, Optional, Type

from sanic_testing.testing import SanicTestClient, TestingResponse  # type: ignore

from pait.app.base import BaseTestHelper
from pait.model.core import PaitCoreModel
from pait.model.response import PaitResponseModel

from ._load_app import load_app

__all__ = ["SanicTestHelper"]


class SanicTestHelper(BaseTestHelper[TestingResponse]):
    client: SanicTestClient

    def _app_init_field(self) -> None:
        if self.cookie_dict:
            self.header_dict.update(self.cookie_dict)

    def _gen_pait_dict(self) -> Dict[str, PaitCoreModel]:
        return load_app(self.client.app)

    @staticmethod
    def _check_resp_status(resp: TestingResponse, response_model: Type[PaitResponseModel]) -> bool:
        return resp.status in response_model.status_code

    @staticmethod
    def _check_resp_media_type(resp: TestingResponse, response_model: Type[PaitResponseModel]) -> bool:
        return resp.content_type == response_model.media_type

    @staticmethod
    def _get_json(resp: TestingResponse) -> dict:
        return resp.json

    def _replace_path(self, path_str: str) -> Optional[str]:
        if self.path_dict and path_str[0] == "<" and path_str[-1] == ">":
            key: str = path_str[1:-1]
            if ":" in key:
                key = key.split(":")[0]
            return self.path_dict[key]
        return None

    def _make_response(self, method: str) -> TestingResponse:
        method = method.lower()
        if method == "get":
            request, resp = self.client._sanic_endpoint_test(method, uri=self.path, headers=self.header_dict)
        else:
            request, resp = self.client._sanic_endpoint_test(
                method,
                uri=self.path,
                data=self.form_dict,
                json=self.body_dict,
                headers=self.header_dict,
                files=self.file_dict,
            )
        return resp



import json
from types import CodeType
from typing import Dict, List

from pydantic.fields import Undefined

from pait import field
from pait.model.core import PaitCoreModel
from pait.model.status import PaitStatus
from pait.util import gen_example_dict_from_schema

from .base_parse import FieldDictType, FieldSchemaTypeDict, PaitBaseParse  # type: ignore


class PaitMd(PaitBaseParse):
    """parse pait dict to md doc"""

    def __init__(
        self,
        pait_dict: Dict[str, PaitCoreModel],
        title: str = "Pait Doc",
        use_html_details: bool = True,
    ):
        """
        :param pait_dict: pait dict
        :param title: Md title
        :param use_html_details: Using HTML syntax-related functions
        """
        self._use_html_details: bool = use_html_details  # some not support markdown in html
        self._title: str = title

        super().__init__(pait_dict)

        self.content = self.gen_markdown_text()
        self._content_type = ".md"

    @staticmethod
    def gen_md_param_table(field_dict_list: List[FieldSchemaTypeDict], blank_num: int = 8) -> str:
        """
        gen param md table
        :param field_dict_list:
        :param blank_num: table str indent
        :return:
            |param name|type|default value|description|other|
            |---|---|---|---|---|
            |age|string|10|user age||
        """
        blank_num_str: str = " " * blank_num
        markdown_text: str = f"{blank_num_str}|param name|type|default value|description|other|\n"
        markdown_text += f"{blank_num_str}|---|---|---|---|---|\n"
        field_dict_list = sorted(field_dict_list, key=lambda x: x["param_name"])
        for field_info_dict in field_dict_list:
            default = field_info_dict["default"]
            if default is Undefined:
                default = "**`Required`**"
            type_ = field_info_dict["type"]
            if default is Undefined:
                type_ = "**`Required`**"
            markdown_text += (
                f"{blank_num_str}"
                f"|{field_info_dict['param_name']}"
                f"|{type_}"
                f"|{default}"
                f"|{field_info_dict['description']}"
                f"|{field_info_dict['other'] or ''}"
                f"|\n"
            )
        return markdown_text

    def gen_markdown_text(self) -> str:
        markdown_text: str = f"# {self._title}\n"
        for group in self._group_list:
            # group
            if self._use_html_details:
                markdown_text += f"<details><summary>Group: {group}</summary>\n\n"
            else:
                markdown_text += f"## Group: {group}\n\n"
            for pait_model in self._group_pait_dict[group]:
                # func info
                markdown_text += f"### Name: {pait_model.operation_id}\n\n"
                status_text: str = ""
                if pait_model.status in (PaitStatus.test, PaitStatus.design, PaitStatus.dev, PaitStatus.integration):
                    status_text = f"<font color=#00BFFF>{pait_model.status.value}</font>"
                elif pait_model.status in (PaitStatus.release, PaitStatus.complete):
                    status_text = f"<font color=#32CD32>{pait_model.status.value}</font>"
                elif pait_model.status in (PaitStatus.abandoned, PaitStatus.abnormal):
                    status_text = f"<font color=#DC143C>{pait_model.status.value}</font>"
                elif pait_model.status:
                    status_text = f"{pait_model.status.value}"

                if pait_model.desc:
                    markdown_text += f"\n\n**Desc**:{pait_model.desc}\n\n"

                # func or interface details
                func_code: CodeType = pait_model.func.__code__  # type: ignore
                markdown_text += "|Author|Status|func|summary|\n"
                markdown_text += "|---|---|---|---|\n"
                markdown_text += (
                    f"|{','.join(pait_model.author) if pait_model.author else ''}"
                    f"|{status_text}"
                    f'|<abbr title="file:{pait_model.func_path or func_code.co_filename};'
                    f'line: {func_code.co_firstlineno}">'
                    f"{pait_model.func.__qualname__}</abbr>"
                    f"|{pait_model.summary}|\n"
                )

                # request info
                markdown_text += f"- Path: {pait_model.path}\n"
                markdown_text += f"- Method: {','.join(pait_model.method_list)}\n"
                markdown_text += "- Request:\n"

                field_dict: FieldDictType = self._parse_func_param_to_field_dict(pait_model.func)
                for pre_depend in pait_model.pre_depend_list:
                    for field_class, field_dict_list in self._parse_func_param_to_field_dict(pre_depend).items():
                        if field_class not in field_dict:
                            field_dict[field_class] = field_dict_list
                        else:
                            field_dict[field_class].extend(field_dict_list)

                # gen key, class can not sort, so replace to instance
                field_key_list: List[field.BaseField] = sorted([i() for i in field_dict.keys()])
                # request body info
                for field_instance in field_key_list:
                    field_class = field_instance.__class__
                    if not issubclass(field_class, field.BaseField):
                        continue
                    field_dict_list = field_dict[field_class]
                    markdown_text += f"{' ' * 4}- {field_class.cls_lower_name().capitalize()} Param\n\n"
                    markdown_text += self.gen_md_param_table(field_dict_list)

                # response info
                markdown_text += "- Response:\n\n"
                if pait_model.response_model_list:
                    for resp_model_class in pait_model.response_model_list:
                        resp_model = resp_model_class()
                        markdown_text += f"{' ' * 4}- {resp_model.name or resp_model.__class__.__name__}\n\n"
                        markdown_text += f"{' ' * 8}|status code|media type|description|\n"
                        markdown_text += f"{' ' * 8}|---|---|---|\n"
                        markdown_text += (
                            f"{' ' * 8}|{','.join([str(i) for i in resp_model.status_code])}"
                            f"|{resp_model.media_type}"
                            f"|{resp_model.description}"
                            f"|\n"
                        )
                        if resp_model.header:
                            markdown_text += f"{' ' * 8}- Header\n"
                            markdown_text += f"{' ' * 12}{resp_model.header}\n"
                        if resp_model.response_data:
                            markdown_text += f"{' ' * 8}- Response Data\n\n"
                            field_dict_list = self._parse_schema(resp_model.response_data.schema())
                            markdown_text += self.gen_md_param_table(field_dict_list, blank_num=12)
                            markdown_text += f"{' ' * 8}- Example Response Json Data\n\n"
                            example_dict: dict = gen_example_dict_from_schema(resp_model.response_data.schema(), False)
                            blank_num_str: str = " " * 12
                            json_str: str = "\n".join(
                                [blank_num_str + i for i in json.dumps(example_dict, indent=2).split("\n")]
                            )
                            markdown_text += f"{blank_num_str}```json\n{json_str}\n{blank_num_str}```\n\n"
                markdown_text += "\n"
            if self._use_html_details:
                markdown_text += "</details>"
        return markdown_text



import inspect
from functools import wraps
from typing import Any, Callable, Dict, List, Optional, Tuple, Type

from pydantic import BaseConfig

from pait.app.base import BaseAppHelper
from pait.g import config, pait_data
from pait.model.core import PaitCoreModel
from pait.model.response import PaitResponseModel
from pait.model.status import PaitStatus
from pait.model.util import sync_config_data_to_pait_core_model
from pait.param_handle import AsyncParamHandler, ParamHandler
from pait.util import get_func_sig


def pait(
    app_helper_class: "Type[BaseAppHelper]",
    make_mock_response_fn: Callable[[Type[PaitResponseModel]], Any],
    enable_mock_response: bool = False,
    pydantic_model_config: Optional[Type[BaseConfig]] = None,
    # param check
    pre_depend_list: Optional[List[Callable]] = None,
    at_most_one_of_list: Optional[List[List[str]]] = None,
    required_by: Optional[Dict[str, List[str]]] = None,
    # doc
    author: Optional[Tuple[str, ...]] = None,
    desc: Optional[str] = None,
    summary: Optional[str] = None,
    name: Optional[str] = None,
    status: Optional[PaitStatus] = None,
    group: Optional[str] = None,
    tag: Optional[Tuple[str, ...]] = None,
    response_model_list: Optional[List[Type[PaitResponseModel]]] = None,
) -> Callable:
    if not isinstance(app_helper_class, type):
        raise TypeError(f"{app_helper_class} must be class")
    if not issubclass(app_helper_class, BaseAppHelper):
        raise TypeError(f"{app_helper_class} must sub from {BaseAppHelper.__class__.__name__}")
    app_name: str = app_helper_class.app_name

    def wrapper(func: Callable) -> Callable:
        # Pre-parsing function signatures
        get_func_sig(func)

        # gen pait core model and register to pait data
        pait_core_model: PaitCoreModel = PaitCoreModel(
            author=author,
            app_helper_class=app_helper_class,
            make_mock_response_fn=make_mock_response_fn,
            desc=desc,
            summary=summary,
            func=func,
            func_name=name,
            status=status,
            group=group,
            tag=tag,
            response_model_list=response_model_list,
            pre_depend_list=pre_depend_list,
            pydantic_model_config=pydantic_model_config,
        )
        sync_config_data_to_pait_core_model(config, pait_core_model, enable_mock_response=enable_mock_response)
        pait_data.register(app_name, pait_core_model)

        if inspect.iscoroutinefunction(func):

            @wraps(func)
            async def dispatch(*args: Any, **kwargs: Any) -> Callable:
                async with AsyncParamHandler(
                    app_helper_class,
                    func,
                    pait_core_model.pydantic_model_config,
                    at_most_one_of_list=at_most_one_of_list,
                    pre_depend_list=pre_depend_list,
                    required_by=required_by,
                    args=args,
                    kwargs=kwargs,
                ) as p:
                    return await pait_core_model.pait_func(*p.args, **p.kwargs)

            return dispatch
        else:

            @wraps(func)
            def dispatch(*args: Any, **kwargs: Any) -> Callable:
                with ParamHandler(
                    app_helper_class,
                    func,
                    pait_core_model.pydantic_model_config,
                    at_most_one_of_list=at_most_one_of_list,
                    pre_depend_list=pre_depend_list,
                    required_by=required_by,
                    args=args,
                    kwargs=kwargs,
                ) as p:
                    return pait_core_model.pait_func(*p.args, **p.kwargs)

            return dispatch

    return wrapper
